#include "stdafx.h"
#include "curve.h"

/** Bezier characteristic matrix */
double BZ[4][4] = {
	{ -1,  3, -3, 1 },
	{  3, -6,  3, 0 },
	{ -3,  3,  0, 0 },
	{  1,  0,  0, 0 }			
};
/* Inverse of Bezier characteristic matrix*/
double iBZ[4][4] = {
	{0,		0,				0,				1},
	{0,		0.0,			0.333333333333333,1},
	{0,		0.333333333333333,0.666666666666666,1},
	{1,		1,				1,				1}
};

/** Catmull-Rom characteristic matrix.  */
double CR[4][4]= {
	{ -0.5,  1.5, -1.5,  0.5 },
	{  1, -2.5,  2, -0.5 },
	{ -0.5,  0,  0.5,  0 },
	{  0,  1,  0,  0 }			
};

CCurve::CCurve(void) : m_finalized(false)
{
}

CCurve::CCurve(CCurve& copy)
{
	m_finalized = copy.m_finalized;
	unsigned int numPoints = copy.m_points.size();
	for (unsigned int i=0;i<numPoints;i++)
	{
		m_points.push_back(copy.m_points[i]);
	}
}

CCurve::~CCurve(void)
{
}

bool CCurve::IsHit(const CPoint& point)
{

	return GetBoundingBox().PtInRect(point)==1?true:false;
}

void CCurve::Draw(CDC& dc)
{
	vector<POINT> points = GetBezierPoints();

	int numPoints = (int)points.size();
	if (numPoints > 0) {
		if (numPoints > 3) {
			POINT* pointArray = new POINT[numPoints];
			for (int j=0;j<numPoints;j++) {
				pointArray[j] = points[j];
			}
			
			dc.PolyBezier(pointArray,numPoints);
			
			delete[] pointArray;
		} else {
			POINT pt0 = points[0];
			dc.MoveTo(pt0.x,pt0.y);
			
			for (int i=1;i<numPoints;i++) {
				POINT pt = points[i];
				
				dc.LineTo(pt.x,pt.y);
				
			}
			if (m_finalized) {
				dc.LineTo(pt0.x,pt0.y);
				
			}
		}
	}
}

void CCurve::DrawDragImage(CDC& dc, CPoint offset)
{
	vector<POINT> points = GetBezierPoints();
	
	int nOldMode = dc.SetROP2(R2_NOT);
	CBrush* pOldBrush =(CBrush*)dc.SelectStockObject(NULL_BRUSH);
	CPen* pen = new CPen(PS_DOT,1,RGB(0,0,0));
	CPen* pOldPen = dc.SelectObject(pen);

	int numPoints = (int)points.size();
	if (numPoints > 0) {
		if (numPoints > 3) {
			POINT* pointArray = new POINT[numPoints];
			for (int j=0;j<numPoints;j++) {
				POINT pt ;
				pt.x = points[j].x + offset.x;
				pt.y = points[j].y + offset.y;

				pointArray[j] = pt;
			} 
			
			dc.PolyBezier(pointArray,numPoints);
			
			delete[] pointArray;
		} else {
			POINT pt0 = points[0];
			pt0.x += offset.x;
			pt0.y += offset.y;
			dc.MoveTo(pt0.x,pt0.y);
			
			for (int i=1;i<numPoints;i++) {
				POINT pt = points[i];
				pt.x += offset.x;
				pt.y += offset.y;
				
				dc.LineTo(pt.x,pt.y);
				
			}
			if (m_finalized) {
				dc.LineTo(pt0.x,pt0.y);
				
			}
		}
	}
	dc.SetROP2(nOldMode);
	dc.SelectObject(pOldBrush);
	dc.SelectObject(pOldPen);

	delete pen;

}


vector<POINT> CCurve::GetBezierPoints() 
{
	vector<POINT> points;

	unsigned int numPoints = m_points.size();
	if (numPoints < 4) {
		return m_points;
	} else {
		m_points.push_back(m_points[numPoints-1]);
		m_points.insert(m_points.begin(),m_points[0]);
		numPoints = m_points.size();
		for (unsigned int i=1;i<numPoints-2;i++) {

			double xVals[4], yVals[4];
			double xResults[4], yResults[4];

			//get four points that make up a CR curve
			for (int j=0;j<4;j++) {
				xVals[j] = m_points[i-1+j].x;
				yVals[j] = m_points[i-1+j].y;
			}

			//convert points to BZ points


			double results[4][4];

			Multiply(iBZ,CR,results);

			Multiply(results,xVals,xResults);
			Multiply(results,yVals,yResults);
			for (int j=0;j<4;j++) {
				if (i==1 || j>0) {
					POINT pt;
					pt.x = static_cast<long>(xResults[j]);
					pt.y = static_cast<long>(yResults[j]);
					points.push_back(pt);
				}
			}
		}
		if (m_finalized) {
			unsigned int n = points.size();
			POINT pt = points[0];
			POINT cp1 = points[n-2];
			cp1.x = 2*points[n-1].x - cp1.x;
			cp1.y = 2*points[n-1].y - cp1.y;
			POINT cp2 = points[1];
			cp2.x = 2*points[0].x - cp2.x;
			cp2.y = 2*points[0].y - cp2.y;
			points.push_back(cp1);
			points.push_back(cp2);
			points.push_back(pt);
		}
		m_points.pop_back();
		m_points.erase(m_points.begin());

		return points;

	}
}

void CCurve::Multiply(double a[4][4], double b[4][4], double result[4][4])
{
	for (int r=0;r<4;r++) 
	{
		for (int c=0;c<4;c++) 
		{
			result[r][c] = 0.0;
			for (int i=0;i<4;i++) 
			{
				result[r][c]+=a[r][i]*b[i][c];
			}
		}
	}
}

void CCurve::Multiply(double matrix[4][4],POINT point, POINT& result)
{
	result.x = static_cast<long>(matrix[0][0] *point.x + matrix[0][1] * point.y + matrix[0][3]);
	result.y = static_cast<long>(matrix[1][0] *point.x + matrix[1][1] * point.y + matrix[1][3]);
}

void CCurve::Multiply(double matrix[4][4], double vals[4], double results[4])
{
	for (int r=0;r<4;r++) 
	{
		results[r] = 0.0;
		for (int c=0;c<4;c++) 
		{
			results[r] += (matrix[r][c] * vals[c]);
		}
	}
}

void CCurve::MoveCurve(CPoint& diff)
{
	int numPoints = GetNumPoints();
	for (int i=0;i<numPoints;i++)
	{
		m_points[i].x+=diff.x;
		m_points[i].y+=diff.y;
	}
}

void CCurve::MovePoint(int index, CPoint& diff)
{
	if (index>=0)
	{
		m_points[index].x+=diff.x;
		m_points[index].y+=diff.y;
	}
}

void CCurve::NormalizeToBoundingBox()
{
	const CRect box = GetBoundingBox();
    
	int numPoints = GetNumPoints();
	for (int i=0;i<numPoints;i++)
	{
		m_points[i].x -= box.left;
		m_points[i].y -= box.top;
	}


}